/**
 * @file   args.c
 * @author Jade Cheng
 * @date   September 28, 2009
 */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "args.h"

/* ------------------------------------------------------------------------- */
extern int args_find(const args_t * args, size_t start, const char * value)
{
        size_t i;

        assert(NULL != args);
        assert(NULL != value);

        /* Loop over each argument, and return the index if found. */
        for (i = start; i < args->count; i++)
                if (0 == strcmp(args->values[i], value))
                        return i;

        /* Return -1 to indicate it is not found. */
        return -1;
}

/* ------------------------------------------------------------------------- */
extern int args_parse(char * line, args_t * args)
{
        char * end;

        assert(NULL != line);
        assert(NULL != args);

        /* Loop over each argument in the command line input. */
        args->count = 0;
        for (;;) {

                /* Skip whitespace. */
                while (0 != isspace(*line))
                        line++;

                /* Check for the end of the line. */
                if (0 == *line)
                        break;

                /* Do not write past the end of the array. */
                if (args->count == ARGS_MAX) {
                        fprintf(stderr, "Too many arguments.\n");
                        return EXIT_FAILURE;
                }

                /* Mark the start of the next argument. */
                args->values[args->count++] = line;

                /* Check for an unquoted argument. */
                if (*line != '"') {

                        /* Search for the next whitespace and overwrite it with
                           a null character to mark the end of the string. */
                        if (NULL != (end = strpbrk(line, " \f\n\r\t\v"))) {
                                *end = 0;
                                line = end + 1;
                                continue;
                        }

                        /* Continue searching from the next character. */
                        line += strlen(line);
                        continue;
                }

                /* Do not consider the '"' character as part of the arg. */
                args->values[args->count - 1] = ++line;

                /* Check for a missing closing '"' character. */
                if (NULL == (end = strchr(line, '"'))) {
                        fprintf(stderr, "Closing '\"' not found.\n");
                        return EXIT_FAILURE;
                }

                /* Mark the end of the string and continue searching from the
                   next character after the closing quotation mark. */
                *end = 0;
                line = end + 1;
        }

        /* Make the args_t structure compatible with the execv family by null
           terminating the last argument. */
        args->values[args->count] = NULL;
        return EXIT_SUCCESS;
}


#ifdef __ARGS_TESTS

/* ------------------------------------------------------------------------- */
int main(int argc, const char * argv[])
{
        args_t myargs;
        size_t i;
        char line[1024];

        argc = argc;
        argv = argv;

        strcpy(line, "sleep 10 &");
        if (0 != args_parse(line, &myargs)) {
                fprintf(stderr, "Failed to parse.\n");
                return EXIT_FAILURE;
        }

        fprintf(stdout, "-------------The string to args_t parssing result-------------\n");
        fprintf(stdout, "The number of values is: %d\n", myargs.count);
        fprintf(stdout, "The argments are:\n");
        for (i = 0; i < myargs.count; i++)
                fprintf(stdout, "%d: [%s]\n", i, myargs.values[i]);

        printf("%d %d %d\n",
                args_find(&myargs, 0, "sleep"),
                args_find(&myargs, 0, "10"),
                args_find(&myargs, 0, "&"));

        return EXIT_SUCCESS;
}

#endif /* __ARGS_TESTS */
