/**
 * @file   ics412sh.c
 * @author Jade Cheng
 * @date   September 28, 2009
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>

#include "args.h"
#include "jobs.h"
#include "pipes.h"

/** A macro to tell gcc an argument is unused intentionally. */
#define UNUSED_ARG(E) E = E

/** The maximum length of a command line. */
#define MAX_INPUT_LENGTH 1024

static void add_jobs(void);
static void display_prompt(void);
static void fork_main(size_t index);
static void handle_cd(const args_t * args);
static void handle_fg(const args_t * args);
static void handle_jobs(const args_t * args);
static void handle_kill(const args_t * args);
static void handle_sigint(int sig_no);
static void start_pids(void);
static void stop_pids(void);
static int support_background(args_t * args);
static int support_redirection(args_t * args);
static void wait_pids(void);

/** A flag indicating the command should execute in the background. */
static int g_background;

/** The number of elements used in the g_pids array. */
static size_t g_pid_count;

/** An array of process ids that are currently executing in the foreground. */
static pid_t g_pids[PIPES_MAX + 1];

/** The parsed pipes structure. */
static pipes_t g_pipes;

/**
 * The name of the file used when redirecting output.  This value is NULL when
 * the output is sent to standard output.
 */
static const char * g_redirect;

/* ------------------------------------------------------------------------- */
/**
 * The main entry point of the program.
 *
 * @param argc The number of command line arguments.
 * @param argv The command line arguments.
 *
 * @return EXIT_SUCCESS or EXIT_FAILURE.
 */
int main(int argc, const char ** argv)
{
        /* Check for usage errors. */
        if (argc != 1) {
                printf("Usage: %s\n", argv[0]);
                return EXIT_SUCCESS;
        }

        /* Intercept SIGINT signals so that the shell can kill its child
           processes instead of being killed itself. */
        if (SIG_ERR == signal(SIGINT, handle_sigint)) {
                perror("Error intercepting SIGINT");
                return EXIT_FAILURE;
        }

        /* Loop over the lifetime of the shell. */
        for (;;) {
                args_t args;
                char   input[MAX_INPUT_LENGTH + 1];
                size_t len;

                /* Reset the global data that depends on the last executed
                   command. */
                g_redirect   = NULL;
                g_background = 0;
                g_pid_count  = 0;
                memset(&g_pipes, 0, sizeof(g_pipes));
                memset(g_pids,   0, sizeof(g_pids));

                /* Display the prompt to the user. */
                display_prompt();

                /* Get a line of input from the user and check for errors. */
                if (NULL == fgets(input, sizeof(input), stdin)) {
                        fprintf(stderr, "Error reading from standard input.\n");
                        return EXIT_FAILURE;
                }

                /* Remove defunct jobs. */
                jobs_flush();

                /* Check for no arguments. */
                len = strlen(input);
                if (len <= 1)
                        continue;

                /* Remove the '\n' from the string returned from fgets. */
                input[len - 1] = 0;

                /* Check for badly-parsed arguments. */
                if (EXIT_SUCCESS != args_parse(input, &args))
                        continue;

                /* Check for the exit and logout commands. */
                if (0 == strcmp(args.values[0], "exit") ||
                    0 == strcmp(args.values[0], "logout"))
                        break;

                /* Check for the command to change directories. */
                if (0 == strcmp(args.values[0], "cd")) {
                        handle_cd(&args);
                        continue;
                }

                /* Check for the command to print the active background jobs. */
                if (0 == strcmp(args.values[0], "jobs")) {
                        handle_jobs(&args);
                        continue;
                }

                /* Check for the command to kill background jobs. */
                if (0 == strcmp(args.values[0], "kill")) {
                        handle_kill(&args);
                        continue;
                }

                /* Check for the command to move a background job to the
                   foreground. */
                if (0 == strcmp(args.values[0], "fg")) {
                        handle_fg(&args);
                        continue;
                }

                /* Support background processing. */
                if (EXIT_SUCCESS != support_background(&args))
                        continue;

                /* Support output redirection. */
                if (EXIT_SUCCESS != support_redirection(&args))
                        continue;

                /* Support pipes. */
                if (EXIT_SUCCESS != pipes_parse(&args, &g_pipes))
                        continue;

                /* Start the processes and continue looping. */
                start_pids();
        }

        /* Arriving here indicates the user typed exit or logout. */
        return EXIT_SUCCESS;
}

/* ------------------------------------------------------------------------- */
/**
 * Adds all jobs from the current command into the background job list.
 */
static void add_jobs(void)
{
        size_t i;

        /* Loop over each process and add it to the job list.  If there is an
           error, give up on the remaining processes. */
        for (i = 0; i < g_pid_count; i++)
                if (EXIT_SUCCESS != jobs_add(g_pids[i]))
                        return;
}

/* ------------------------------------------------------------------------- */
/**
 * Displays the prompt to the user.
 */
static void display_prompt(void)
{
        /* Display the prompt and flush the output.  Note printf does not
           necessarily flush when there is no '\n' character. */
        printf("ics412sh%% ");
        fflush(stdout);
}

/* ------------------------------------------------------------------------- */
/**
 * The main entry point for forked, child processes.
 *
 * @param The child's corresponding index in the g_pipes pipes and fd arrays.
 */
static void fork_main(size_t child_index)
{
        size_t i;

        /* Redirect the standard input for piped targets. */
        if (child_index > 0) {
                if (-1 == dup2(g_pipes.fd[child_index - 1].read, 0)) {
                        perror("Error duplicating read file descriptor");
                        exit(EXIT_FAILURE);
                }
        }

        /* Redirect the standard output for piped sources. */
        if (child_index < g_pipes.count) {
                if (-1 == dup2(g_pipes.fd[child_index].write, 1)) {
                        perror("Error duplicating write file descriptor");
                        exit(EXIT_FAILURE);
                }
        }

        /* Close file descriptors for pipes not needed by this child. */
        for (i = 0; i < g_pipes.count; i++) {
                if (i != child_index)
                        close(g_pipes.fd[i].write);
                if (i + 1 != child_index)
                        close(g_pipes.fd[i].read);
        }

        /* Check to redirect output.  In this case, create or open and truncate
           the output file and redirect output to this file. */
        if (child_index == g_pipes.count && g_redirect != NULL) {
                int fd;

                /* Create the file with 644 permissions. */
                if (-1 == (fd = open(g_redirect, O_WRONLY | O_CREAT | O_TRUNC,
                                     0644))) {
                        fprintf(stderr, "Error creating file '%s': %s\n",
                                g_redirect, strerror(errno));
                        exit(EXIT_FAILURE);
                }

                /* Redirect standard output to this file descriptor. */
                if (-1 == dup2(fd, 1)) {
                        perror("Error duplicating write file descriptor");
                        exit(EXIT_FAILURE);
                }
        }

        /* Start the process and check for errors.  Note this function does not
           return if it is successful. */
        if (0 != execvp(g_pipes.children[child_index].values[0],
                        g_pipes.children[child_index].values)) {
                fprintf(stderr, "Error executing command '%s': %s\n",
                        g_pipes.children[child_index].values[0], strerror(errno));
                exit(EXIT_FAILURE);
        }

        /* Either execvp or exit should occur, so execution should not reach
           this point. */
        assert(0);
}

/* ------------------------------------------------------------------------- */
/**
 * Changes the current directory.
 *
 * @param args The parsed command line arguments.
 */
static void handle_cd(const args_t * args)
{
        char * directory;

        /* Check for usage errors. */
        if (args->count > 2) {
                fprintf(stderr, "Usage: cd [directory]\n");
                return;
        }

        /* Check if a directory is specified.  If not, use the $HOME environment
           variable.  If $HOME is not set, then give up. */
        if (args->count == 2) {
                directory = args->values[1];
        } else if (NULL == (directory = getenv("HOME"))) {
                fprintf(stderr, "HOME enviroment variable is not set.\n");
                return;
        }

        /* Change directories and check for errors. */
        if (0 != chdir(directory)) {
                perror(NULL);
                return;
        }
}

/* ------------------------------------------------------------------------- */
/**
 * Handle the "fg" command to move a job into the foreground.
 *
 * @param The parsed command line arguments.
 */
static void handle_fg(const args_t * args)
{
        int job;

        /* Check for usage errors. */
        if (args->count != 2) {
                fprintf(stderr, "Usage: fg job-index\n");
                return;
        }

        /* Read the job index as as integer. */
        if (1 != sscanf(args->values[1], "%d", &job) || job < 0) {
                fprintf(stderr, "Invalid job index: %s.\n", args->values[1]);
                return;
        }

        /* Move the job into the foreground.  This function does not return
           until the job terminates or an error has occurred. */
        jobs_fg((size_t)job);
}

/* ------------------------------------------------------------------------- */
/**
 * Handle the "jobs" command, which prints the active background processes.
 *
 * @param args The parsed command line arguments.
 */
static void handle_jobs(const args_t * args)
{
        /* Check for usage errors. */
        if (args->count != 1) {
                fprintf(stderr, "Usage: jobs\n");
                return;
        }

        /* Print each active background process and its index. */
        jobs_print();
}

/* ------------------------------------------------------------------------- */
/**
 * Handles the "kill" command, which kills a background process.
 *
 * @param The parsed command line arguments.
 */
static void handle_kill(const args_t * args)
{
        int job;

        /* Check for usage errors. */
        if (args->count != 2) {
                fprintf(stderr, "Usage: kill job-index\n");
                return;
        }

        /* Read the job index as as integer. */
        if (1 != sscanf(args->values[1], "%d", &job) || job < 0) {
                fprintf(stderr, "Invalid job index: %s.\n", args->values[1]);
                return;
        }

        /* Kill the job if it exists.  This function does not return until the
           job terminates or an error has occurred. */
        jobs_kill((size_t)job);
}

/* ------------------------------------------------------------------------- */
/**
 * Handles the SIGINT signal.  This occurs when the user presses Control-C.
 * This functino will kill the actively executing command if one exists.
 *
 * @param unused This argument is unused.
 */
static void handle_sigint(int unused)
{
        UNUSED_ARG(unused);

        /* Just redisplay the prompt if nothing is executing in the
           foreground. */
        if (0 == g_pid_count) {
                printf("\n");
                display_prompt();
                return;
        }

        /* Stop all processes executing in the foreground. */
        stop_pids();
}

/* ------------------------------------------------------------------------- */
/**
 * Stops all processes executing in the foreground.
 */
static void stop_pids(void)
{
        size_t i;

        /* Loop over all processes executing in the foreground.  Kill the
           process, checking for errors.  If an error occurs, display a message
           but keep looping of the remaining processes. */
        for (i = 0; i < g_pid_count; i++)
                if (0 != g_pids[i] && 0 != kill(g_pids[i], SIGINT))
                        fprintf(stderr,
                                "Error sending SIGINT to pid %ld: %s\n",
                                (long)g_pids[i], strerror(errno));

        /* There are no more processes running in the foreground. */
        g_pid_count = 0;

        /* Close all open pipes. */
        for (i = 0; i < g_pipes.count; i++) {
                close(g_pipes.fd[i].read);
                close(g_pipes.fd[i].write);
        }

        /* There are no more active pipes. */
        g_pipes.count = 0;
}

/* ------------------------------------------------------------------------- */
/**
 * Starts the processes specified by the user from the command line.
 */
static void start_pids(void)
{
        size_t i;

        /* Loop over each process.  Note there is always one more processes than
           pipes; i.e. if there are no pipes, there is only one process. */
        for (i = 0; i <= g_pipes.count; i++) {
                pid_t pid;

                /* Fork, and check for errors.  In the case of an error, kill
                   all other processes immediately. */
                if (-1 == (pid = fork())) {
                        stop_pids();
                        return;
                }

                /* Check if this is the main process.  If it is, record this
                   process id and continue looping. */
                if (0 != pid) {
                        g_pids[i] = pid;
                        g_pid_count++;
                        continue;
                }

                /* Execute the main method for the forked process.  This
                   function does not return. */
                fork_main(i);
                assert(0);
        }

        /* The main process does not use any of the pipes directly, so close
           them here.  Note that the pipes are reference counted, so each
           process must close them before the pipe resources are removed from
           the system. */
        for (i = 0; i < g_pipes.count; i++) {
                close(g_pipes.fd[i].read);
                close(g_pipes.fd[i].write);
        }

        /* Either launch the processes in the foreground or the background. */
        if (0 == g_background)
                wait_pids();
        else
                add_jobs();
}

/* ------------------------------------------------------------------------- */
/**
 * Processes the parsed command line arguments to support background processes
 * (specified by the '&' character).
 *
 * @param The parsed command line arguments.
 *
 * @return EXIT_SUCCESS or EXIT_FAILURE
 */
static int support_background(args_t * args)
{
        int pos;

        /* Check for an '&' character.  If there is no character, the command
           will execute in the foreground.  This is okay, so return
           successfully. */
        if (-1 == (pos = args_find(args, 0, "&")))
                return EXIT_SUCCESS;

        /* Check that the '&' character is the last character on the command
           line.  If it isn't then display an error and return a failure. */
        if (pos + 1 != (int)args->count) {
                fprintf(stderr, "Invalid background process specifier.\n");
                return EXIT_FAILURE;
        }

        /* Store a flag indicating the command will execute in the background,
           and remove this '&' character from the array of arguments passed to
           the child processes. */
        g_background = 1;
        args->values[pos] = NULL;
        args->count--;
        return EXIT_SUCCESS;
}

/* ------------------------------------------------------------------------- */
/**
 * Processes the parsed command line arguments to support redirection of
 * standard output (specified by the '>' character).
 *
 * @param The parsed command line arguments.
 *
 * @return EXIT_SUCCESS or EXIT_FAILURE
 */
static int support_redirection(args_t * args)
{
        int pos;

        /* Check for a '>' character.  If there is no character, the command
           output will not be redirected.  This is okay, so return
           successfully. */
        if (-1 == (pos = args_find(args, 0, ">")))
                return EXIT_SUCCESS;

        /* Check that the '>' character is the second-to-last character on the
           command line.  If it isn't then display an error and return a
           failure. */
        if (pos + 2 != (int)args->count) {
                fprintf(stderr, "Invalid redirection.\n");
                return EXIT_FAILURE;
        }

        /* Also check this is the only redirection character.  This is a case
           that isn't supported by this shell. */
        if (-1 != args_find(args, pos + 1, ">")) {
                fprintf(stderr, "Invalid redirection.\n");
                return EXIT_FAILURE;
        }

        /* Store the name of the file that will receive the standard output
           from the command, and remove the '>' character from the arguments
           so that it is not processes further. */
        g_redirect = args->values[args->count - 1];
        args->values[pos] = NULL;
        args->count -= 2;
        return EXIT_SUCCESS;
}

/* ------------------------------------------------------------------------- */
/**
 * Waits for the list of processes to complete.
 */
static void wait_pids(void)
{
        size_t i;

        /* Loop over all processes in the g_pids array.  These are the processes
           executing in the foreground. */
        for (i = 0; i < g_pid_count; i++) {
                pid_t pid;
                int   status;

                /* Remove the pid from the array. */
                pid = g_pids[i];
                g_pids[i] = 0;

                /* Wait for the process to terminate.  If this is successful,
                   continue looping. */
                if (-1 != waitpid(pid, &status, 0))
                        continue;

                /* Display an error and abort execution of the remaining
                   processes.  This usually indicates a broken pipe. */
                fprintf(stderr, "Error waiting for pid %ld: %s\n", (long)pid,
                        strerror(errno));
                stop_pids();
                return;
        }
}
