/**
 * @file   jobs.c
 * @author Jade Cheng
 * @date   September 28, 2009
 */

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "jobs.h"

/** The maximum number of background jobs that can execute at one time. */
#define JOBS_MAX 1024

/** The job list. */
static int g_jobs[JOBS_MAX];

/* ------------------------------------------------------------------------- */
extern int jobs_add(pid_t pid)
{
        size_t i;

        assert(pid > 0);

        /* Look for an unused element in the job list, and store this job
           there.  Print a message and return successfully when the job is
           stored. */
        for (i = 0; i < JOBS_MAX; i++) {
                if (g_jobs[i] == 0) {
                        printf("job [%d]\n", i + 1);
                        g_jobs[i] = pid;
                        return EXIT_SUCCESS;
                }
        }

        /* Display an error message and fail if there are no more available
           elements in the job list array. */
        fprintf(stderr, "Too many background jobs.\n");
        return EXIT_FAILURE;
}

/* ------------------------------------------------------------------------- */
extern void jobs_kill(size_t job_index)
{
        pid_t pid;

        /* Check for an invalid job index, and check if there is no job at that
           index.  The index is one greater than the array index. */
        if (job_index == 0 ||
            job_index >= JOBS_MAX ||
           0 == (pid = g_jobs[job_index - 1])) {
                fprintf(stderr, "Invalid job number: %d\n", job_index);
                return;
        }

        /* Clear the job from the job list, and send it the SIGINT signal. */
        g_jobs[job_index - 1] = 0;
        if (0 != kill(pid, SIGINT)) {
                fprintf(stderr, "Error sending SIGINT to pid %ld: %s\n",
                        (long)pid, strerror(errno));
        }
}

/* ------------------------------------------------------------------------- */
extern void jobs_flush(void)
{
        size_t i;

        /* Loop over each job in the job list. */
        for (i = 0; i < JOBS_MAX; i++) {
                pid_t pid;
                int   status;

                /* Skip unused elements. */
                if (g_jobs[i] == 0)
                        continue;

                /* Check if the process is still running and continue if it
                   still is.  Note the waitpid function usually blocks until the
                   process terminates, but in this case the WNOHANG flag
                   indicates the function should return immediately if the
                   process is still running. */
                if (0 == (pid = waitpid(g_jobs[i], &status, WNOHANG)))
                        continue;

                /* Clear the job element, and display a message. */
                g_jobs[i] = 0;
                if (pid > 0) {
                        printf("[%d] Done\n", i + 1);
                } else {
                        fprintf(stderr, "[%d] Failed: %s\n", i + 1,
                                strerror(errno));
                }
        }
}

/* ------------------------------------------------------------------------- */
extern void jobs_fg(size_t job_index)
{
        pid_t pid;
        int status;

        /* Check for an invalid job index, and check if there is no job at that
           index.  The index is one greater than the array index. */
        if (job_index == 0 || 
            job_index >= JOBS_MAX || 
            0 == (pid = g_jobs[job_index - 1])) {
                fprintf(stderr, "Invalid job number: %d\n", job_index);
                return;
        }

        /* Clear the job element and wait for the job to complete.  Note the
           waitpid function blocks until the process terminates or otherwise
           fails. */
        g_jobs[job_index - 1] = 0;
        if (-1 == waitpid(pid, &status, 0)) {
                fprintf(stderr, "Error waiting for pid %ld: %s\n",
                        (long)pid, strerror(errno));
        }
}

/* ------------------------------------------------------------------------- */
extern void jobs_print(void)
{
        size_t i;

        /* Loop over the global array, check if a job is running, and display
           its job index if it is.  The job index is one greater than the
           array index. */
        for (i = 0; i < JOBS_MAX; i++)
                if (g_jobs[i] != 0)
                        printf("[%d] Running\n", i + 1);
}


#ifdef __JOBS_TESTS

/* ------------------------------------------------------------------------- */
int main(int argc, const char * argv[])
{
        argc = argc;
        argv = argv;

        return EXIT_SUCCESS;
}

#endif /* __JOBS_TESTS */
