/**
 * @file   pipes.c
 * @author Jade Cheng
 * @date   September 28, 2009
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "args.h"
#include "pipes.h"

static void copy_args(
        args_t       * dst,
        const args_t * src,
        size_t         src_index,
        size_t         count);

static int create_file_descriptors(pipes_t * pipes);

/* ------------------------------------------------------------------------- */
extern int pipes_parse(const args_t * args, pipes_t * pipes)
{
        size_t i, begin;

        assert(NULL != args);
        assert(NULL != pipes);
        assert(args->count > 0);

        /* Assume there are no pipes found (this is okay). */
        pipes->count = 0;

        /* Check for a leading pipe, and return an error if one is found. */
        if (0 == strcmp(args->values[0], "|")) {
                fprintf(stderr, "Invalid arguments: Leading pipe found.\n");
                return EXIT_FAILURE;
        }

        /* Check for a trailing pipe, and return an error if one is found. */
        if (0 == strcmp(args->values[args->count - 1], "|")) {
                fprintf(stderr, "Invalid arguments: Trailing pipe found.\n");
                return EXIT_FAILURE;
        }

        /* Look for pipes, starting with the first argument. */
        begin = 0;
        for (i = 0; i < args->count; i++) {

                /* Skip this argument if it is not a pipe. */
                if (0 != strcmp(args->values[i], "|"))
                        continue;

                /* Check for adjacent pipes. */
                if (begin == i) {
                        fprintf(stderr, "Invalid arguments: Adjacent pipes.\n");
                        return EXIT_FAILURE;
                }

                /* Check for too many pipes. */
                if (pipes->count == PIPES_MAX) {
                        fprintf(stderr, "Invalid arguments: Too many pipes.\n");
                        return EXIT_FAILURE;
                }

                /* Copy the arguments from the args_t structure to the pipes_t
                   structure. */
                copy_args(
                        &pipes->children[pipes->count++],
                        args,
                        begin,
                        i - begin);

                /* In the next iteration, look from the following argument. */
                begin = i + 1;
        }

        /* Copy the remaining arguments into the pipes_t structure. */
        copy_args(
                &pipes->children[pipes->count],
                args,
                begin,
                args->count - begin);

        /* Create file descriptors for each pipe. */
        return create_file_descriptors(pipes);
}

/* ------------------------------------------------------------------------- */
/**
 * Copies data from one args_t structure to another.  This function writes NULL
 * to the value at the end of the destination structure values array.  This
 * makes it safe to call from the execv family of functions.
 *
 * @param dst The destination arguments structure.
 * @param src The source arguments structure.
 * @param src_index The index of the first argument in the source structure.
 * @param count The number of arguments to copy.
 */
static void copy_args(
        args_t       * dst,
        const args_t * src,
        size_t         src_index,
        size_t         count)
{
        assert(NULL != dst);
        assert(NULL != src);
        assert(src_index + count <= ARGS_MAX);

        /* Store the count, the values, and NULL terminate the array. */
        dst->count = count;
        memcpy(dst->values, &src->values[src_index], sizeof(char *) * count);
        dst->values[count] = NULL;
}

/* ------------------------------------------------------------------------- */
/**
 * Creates file descriptors for each pipe in a pipes_t structure.
 *
 * @param pipes The pipes structure.
 *
 * @return EXIT_SUCCESS or EXIT_FAILURE.
 */
static int create_file_descriptors(pipes_t * pipes)
{
        size_t i;

        assert(NULL != pipes);

        /* Loop over each pipe in the structure. */
        for (i = 0; i < pipes->count; i++) {
                int fd[2];

                /* Check if creating the pipe fails. */
                if (-1 == pipe(fd)) {
                        size_t j;

                        /* Display an error message. */
                        perror("Error creating pipes");

                        /* Delete all pipes previously created. */
                        for (j = 0; j < i; j++) {
                                close(pipes->fd[j].read);
                                close(pipes->fd[j].write);
                        }

                        /* Do not attempt to create any more pipes. */
                        return EXIT_FAILURE;
                }

                /* Store the read and write file descriptors into the pipes
                   structure. */
                pipes->fd[i].read  = fd[0];
                pipes->fd[i].write = fd[1];
        }

        /* Return successfully. */
        return EXIT_SUCCESS;
}

#ifdef __PIPES_TESTS

/* ------------------------------------------------------------------------- */
int main(int argc, const char * argv[])
{
        args_t  myargs;
        pipes_t mypipes;
        size_t  i, j;
        char    line[1024];

        argc = argc;
        argv = argv;

        strcpy(line, "a b");
        if (0 != args_parse(line, &myargs)) {
                fprintf(stderr, "Failed to parse.\n");
                return EXIT_FAILURE;
        }

        fprintf(stdout, "-------The string to args_t parssing result-------\n");
        fprintf(stdout, "The number of values is: %d\n", myargs.count);
        fprintf(stdout, "The argments are:\n");
        for (i = 0; i < myargs.count; i++)
                fprintf(stdout, "%d: [%s]\n", i, myargs.values[i]);

        if (0 != pipes_parse(&myargs, &mypipes)) {
                fprintf(stderr, "Error passing args_t to pipes_t.\n");
                return EXIT_FAILURE;
        }

        fprintf(stdout, "------The args_t to pipes_t parssing result------\n");
        fprintf(stdout, "The number of pipes is: %d\n", mypipes.count);
        for (i = 0; i <= mypipes.count; i++) {
                fprintf(stdout, "the number %d child has %d values.\n",
                        i, mypipes.children[i].count);
                for (j = 0; j < mypipes.children[i].count; j++)
                        fprintf(stdout, "%d: [%s]\n", j,
                                mypipes.children[i].values[j]);
        }

        return EXIT_SUCCESS;
}

#endif /* __PIPES_TESTS */
